//********************************************************************
//*** ResourcePool/ResourcePool.hh
//*** Copyright (c) 2003-2005 by Markus Winand <mws@fatalmind.com>
//*** $Id: ResourcePool.cc,v 1.22 2005/09/20 11:47:16 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef RESOURCEPOOL_RESOURCEPOOL_HH
#include "ResourcePool/ResourcePool.hh"
#endif

#ifndef RESOURCEPOOL_CREATEEXCEPTION_HH
#include "CreateException.hh"
#endif

//#include <iostream>

namespace fatalmind {

template<class Type>
ResourcePool<Type>::ResourcePool(
	  const typename Type::factory_t& a_factory
	, const ResourcePoolOptions& a_options
): _RPExecute<Type>()
	, _factory(a_factory)
	, _options(a_options)
	, _stopped(false)
	//, CommandFactory(_factory)
{
	try {
		for (int i = _options.getPreCreate(); i > 0; --i) {
			_inc_pool();
		}
	} catch (CreateException e) {
	}
}

template<class Type>
ResourcePool<Type>::~ResourcePool()
{
	sync_t lk(_condvar);
	_stopped = true;
	_downsize();
	if (TM::isMultiThreaded) {
		while (!_usedPool.empty()) {
			_condvar.wait();
			_downsize();
		}
	}
}

template<class Type>
void ResourcePool<Type>::_downsize()
{
	typename freePool_t::iterator i = _freePool.begin();
	typename freePool_t::iterator end = _freePool.end();
	for (; i != end; ++i) {
		(*i)->close();
	}
	_freePool.clear();
}

template<class Type>
void ResourcePool<Type>::downsize()
{
	sync_t lk(_condvar);
	_downsize();
}

template<class Type>
ResourcePoolStats ResourcePool<Type>::getStats() const
{
	sync_t lk(_condvar);
	// should be safe, i guess, ISO C++ 12.2
	return ResourcePoolStats(_stats, _usedPool.size(), _freePool.size()); 
}

template<class Type>
typename Type::plainresource_t& ResourcePool<Type>::get()
{
	typename Type::resource_t* rec = 0;
	int maxtry = _options.getMaxTry() - 1;
	CreateException orig(
		"There is currently no valid resource to be used"
	);

	do {
		try {
			rec = moveToUsed();
			if (! rec->precheck()) {
				sync_t lk(_condvar);
				rec->fail_close();
				_usedPool.erase(plainresource_ptr(rec->get_plain_resource()));
				rec = 0;
			}
		} catch (CreateException e) {
			orig = e;	
		}
	} while ((rec == 0) && (maxtry-- > 0) && (sleepit(maxtry)));

	if (rec != 0) {
		return rec->get_plain_resource();
	} else {
		throw orig;
	}
}

// returns direct reference
template<class Type>
typename Type::resource_t* ResourcePool<Type>::moveToUsed()
{
	sync_t lk(_condvar);
	_stats.getCall_();

	if (_stopped) {
		throw CreateException("This ResourcePool is beeing stopped");
	}
	// TODO: make noWait() and Wait(time) in options
	while (_freePool.empty()) {
		try {
			_inc_pool();
		} catch (CreateException e) {
			_stats.waitEvent_();
			_waitForFree(e, typename TM::Type());
		}
	}
	
	resource_ptr rec(_freePool.front());
	_freePool.pop_front();
	_usedPool[plainresource_ptr(rec->get_plain_resource())] = rec;

	return &*rec;
}

template<class Type>
void ResourcePool<Type>::_waitForFree(
	  const CreateException& e
	, const SingleThreadedModel<>::Type t
) {
	throw e;	
}

template<class Type>
void ResourcePool<Type>::_waitForFree(
	  const CreateException& e
	, const MultiThreadedModel<>::Type t
) {
	if (! _condvar.wait(_options.getWait())) {
		throw e;
	}	
}

template<class Type>
bool ResourcePool<Type>::free(const typename Type::plainresource_t& plain_res) 
{
	sync_t lk(_condvar);
	typename usedPool_t::iterator i 
		= _usedPool.find(plainresource_ptr(plain_res));
	if (i != _usedPool.end()) {
		resource_ptr res = i->second;
		_usedPool.erase(i);	
		if (res->postcheck()) {
			_freePool.push_back(res);
			_condvar.signal();
		} else {
			res->fail_close();
		}
		return true;
	}
	return false;
}

template<class Type>
bool ResourcePool<Type>::fail(const typename Type::plainresource_t& plain_res)
{
	sync_t lk(_condvar);
	_stats.failCall_();

	typename usedPool_t::iterator i 
		= _usedPool.find(plainresource_ptr(plain_res));
	if (i != _usedPool.end()) {
		resource_ptr res = i->second;
		_usedPool.erase(i);
		res->fail_close();
		return true;
	}
	return false;
}

template<class Type>
void ResourcePool<Type>::_inc_pool()
{
	int PoolSize = _freePool.size() + _usedPool.size();

	if (PoolSize < _options.getMax()) {
		resource_ptr res(_factory.create_resource());
		_freePool.push_back(res);
	} else {
		throw CreateException(
			"No more resource available because the Max value has been reached"
		);
	}
}

template<class Type>
bool ResourcePool<Type>::sleepit(int tryno) const
{
	return _options.SleepOnFail(tryno);
}


} // namespace fatalmind
