//********************************************************************
//*** ResourcePool/ResourcePool.t.cc
//*** Copyright (c) 2003-2005 by Markus Winand <mws@fatalmind.com>
//*** $Id: ResourcePool.t.cc,v 1.25 2005/09/17 14:01:05 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#include "Execute.cc"
#include "ResourcePool.hh"
#include "ResourcePool.cc"
#include "Command.cc"

#include "Thread/Thread.hh"

#include "../t/main.hh"

#include <string>
#include <sstream>

#include "POSIX/sys/timeval.hh"
#include <sys/time.h>

using namespace std;
using namespace fatalmind;

class plain {
	public:
		plain(string& arg)
		: _arg(arg)
		, _ok(true)
		{
		}

		bool ok() {
			return _ok;
		}

		void ok(bool ok) {
			_ok = ok;
		}

		string& str()
		{
			return _arg;
		}

		bool operator<(const plain& rho) const {
			return _arg < rho._arg;
		}
	private:
		string _arg;
		bool	_ok;
};

class Rec {
	public:
		Rec(const string& arg) 
		: _arg(arg)
		, _pl(_arg)
		{
		}
		bool precheck() {
			return _pl.ok();
		};
		bool postcheck() {
			return _pl.ok();
		};
		void close() {
		};
		void fail_close() {
		};

		plain& get_plain_resource() {
			return _pl;
		};
	private:
		string _arg;
		plain  _pl;
};

class Fac {
	public:
		mutable int	_cnt;
		string _arg;
	public:
		Fac(char* arg) 
		: _cnt(0)
		, _arg(arg)
		{
		};
		Rec* create_resource() const 
		{
			_cnt++;
			ostringstream os;
			os << _arg << "(" << _cnt << ")";
			return new Rec(os.str());
		};
};

template<class TM>
class myType: public PoolType<Fac, Rec, plain, NopCommand<Fac>, TM> {
};
template<class DTM = DefaultThreadedModel>
class RP: public fatalmind::ResourcePool<myType<DTM> >
{
	typedef fatalmind::ResourcePool<myType<DTM> > super;
	public:
		RP(const typename myType<DTM>::factory_t& f, const ResourcePoolOptions& opti = ResourcePoolOptions())
		: super(f, opti)
		{
		};

		using super::getFactory; // make getFactory public
};

#ifndef THIS_IS_A_VERY_STRANGE_INCLUDE

void TEST_plain() {
	Fac F("hirsch");
	RP<> x(F);
	plain& res = x.get();
	assertEqualT(string, res.str(), "hirsch(1)");
	assertEqualT(bool, x.free(res), true);
	assertEqualT(bool, x.fail(res), false);	
}

void TEST_fail() {
	Fac F("hirsch");
	RP<> x(F);
	plain& res = x.get();
	x.free(res);

	{
		plain& res2 = x.get();
		x.free(res2);
		assertEqualT(string, res2.str(), "hirsch(1)");
	}
	

	{
		plain& res2 = x.get();
		x.fail(res2);
	}
	{
		plain& res2 = x.get();
		x.free(res2);
		assertEqualT(string, res2.str(), "hirsch(2)");
	}
}

void TEST_precheck() {
	Fac F("hirsch");
	RP<> x(F);
	plain& rec = x.get();
	x.free(rec);
	rec.ok(false);
	plain& rec2 = x.get();
	string result = rec2.str();
	x.free(rec2);
	assertEqualT(string, result, "hirsch(2)");
}

void TEST_postcheck() {
	Fac F("hirsch");
	RP<> x(F);
	plain& rec = x.get();
	rec.ok(false);
	x.free(rec);
	{
		plain& rec = x.get();
		string result = rec.str();
		x.free(rec);
		assertEqualT(string, result, "hirsch(2)");
	}
}

void TEST_Options() {
	Fac f1("f1");
	Fac f2("f2");
	Fac f3("f3");
	Fac f4("f4");
	Fac f5("f5");

	ResourcePoolOptions o1;
	o1.setPreCreate(3);
	o1.setWait(0);

	ResourcePoolOptions o2;
	o2.setMax(1);
	o2.setSleepOnFail(1);
	o2.setWait(0);

	ResourcePoolOptions o3;
	o3.setMax(1);
	o3.setPreCreate(1);
	o3.setWait(0);

	ResourcePoolOptions o4;
	o4.setMax(1);
	o4.setMaxTry(4);
	o4.setSleepOnFail(0, 1, 2);
	o4.setWait(0);

	ResourcePoolOptions o5;
	o5.setMax(1);
	o5.setMaxTry(3);
	o5.setSleepOnFail(0, 1, 2);
	o5.setPreCreate(10);
	o5.setWait(0);

	RP<> p1(f1, o1);
	::POSIX::timeval start, stop;
	struct timezone tz;
	gettimeofday(&start, &tz);

	assertEqualT(int, p1.getFactory()._cnt, 3);
	plain& r1 = p1.get();
	plain& r2 = p1.get();
	plain& r3 = p1.get();
	plain& r4 = p1.get();
	plain& r5 = p1.get();
	assertEqualT(int, p1.getFactory()._cnt, 5);
	try {
		plain& r6 = p1.get();
		assertNotReach();
		p1.fail(r6);
	} catch (CreateException e) {
	}
	p1.free(r1);
	p1.free(r2);
	p1.free(r3);
	p1.free(r4);
	p1.free(r5);
	gettimeofday(&stop, &tz);
	assertLessT(float, stop-start, 0.5);

	// Pool 2
	{
	RP<> p2(f2, o2);
	assertEqualT(int, p2.getFactory()._cnt, 0);
	plain& r1 = p2.get();
	assertEqualT(int, p2.getFactory()._cnt, 1);
	gettimeofday(&start, &tz);

	try {
		plain& r2 = p2.get();
		assertNotReach();
		p2.free(r2);
	} catch (CreateException e) {
	}
	gettimeofday(&stop, &tz);
	p2.free(r1);
	assertLessT(float, stop-start, 1.5);
	assertMoreT(float, stop-start, 0.5);
	
	}
	// Pool 3
	{
	RP<> p3(f3, o3);
	assertEqualT(int, p3.getFactory()._cnt, 1);
	gettimeofday(&start, &tz);
	plain& r6 = p3.get();
	try {
		plain& r2 = p3.get();
		assertNotReach();
		p3.free(r2);
	} catch (CreateException e) {
	}
	assertEqual(p3.free(r6), true);
	plain &r3 = p3.get();

	gettimeofday(&stop, &tz);
	assertLessT(float, stop-start, 0.5);	
	p3.free(r3);

	}
	// pool 4
	{
	RP<> p4(f4, o4);
	gettimeofday(&start, &tz);

	plain& r1 = p4.get();
    try {
        plain& r2 = p4.get();
        assertNotReach();
		p4.free(r2);
    } catch (CreateException e) {
    }

	gettimeofday(&stop, &tz);
	assertLessT(float, stop-start, 3.5);
	assertMoreT(float, stop-start, 2.5);
	p4.free(r1);

	}
	// pool 5
	{
	RP<> p5(f5, o5);
	assertEqualT(int, p5.getFactory()._cnt, 1);
	gettimeofday(&start, &tz);

	plain& r1 = p5.get();	
	try {
		plain& r2 = p5.get();
        assertNotReach();
		p5.free(r2);
	} catch (CreateException e) {
	}

	gettimeofday(&stop, &tz);

    assertLessT(float, stop-start, 1.5);
    assertMoreT(float, stop-start, 0.5);
	p5.free(r1);
	}
}

template<class TM>
class MyTestTask : public Task
{
	public:
		MyTestTask(RP<TM>& p, int limit) 
		: p_(p)
		, limit_(limit)
		{
		}

		void run() {
			for (int i = 0; i < limit_; ++i) {
				plain& res = p_.get();
				p_.free(res);
			}
		}
	private:
		RP<TM>& p_;
		int     limit_;
};

Fac F("hirsch");
RP<DefaultThreadedModel> rp(F);

bool perf_tpl(int p) {
	for (int i = 0; i < p; ++i) {
		plain& res = rp.get();
		rp.free(res);
	}

	return true;
}

void TEST_PERF_MULTI_simple()
{
	perf_tpl(10000);
}

#endif
